import json
import ssl
import traceback
from datetime import datetime
from itertools import chain

import requests
from requests.adapters import HTTPAdapter

from tornado.log import app_log


class SSLAdapter(HTTPAdapter):
    def __init__(self, ssl_context=None, **kwargs):
        self.ssl_context = ssl_context
        super().__init__(**kwargs)

    def init_poolmanager(self, *args, **kwargs):
        kwargs["ssl_context"] = self.ssl_context
        return super().init_poolmanager(*args, **kwargs)


class CloudClient:
    def __init__(self, base_url, ciphers=None):
        self.base_url = base_url
        self.ciphers = (
            ciphers
            if ciphers
            else "ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-GCM-SHA256:ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384:DHE-RSA-AES128-GCM-SHA256:DHE-RSA-AES256-GCM-SHA384"
        )

    def get_context(self):
        context = ssl.create_default_context()
        context.set_ciphers(self.ciphers)
        return context

    def request(self, url, body=None, headers=None):
        # 创建异步请求
        base_headers = {"Content-Type": "application/json"}
        if headers:
            base_headers.update(headers)

        session = requests.Session()
        session.mount("https://", SSLAdapter(ssl_context=self.get_context()))

        response = session.post(url, json=body, headers=headers, timeout=5)
        if response.status_code != 200:
            raise ValueError(f"访问{url}接口失败")
        return json.loads(response.content.decode("utf-8"))

    def get_token(self, ems_sn, sn_secret):
        token_url = self.base_url + "/device/token/getByRegistrationSn"
        try:
            response_data = self.request(
                url=token_url,
                body={
                    "registrationSn": ems_sn,
                    "snSecret": sn_secret,
                },
            )
            app_log.info(f"获取token结果 {response_data}")
            if response_data.get("result"):
                return response_data["result"]
        except Exception as e:
            app_log.info(f"访问token接口失败: {str(e)}")

    def get_weather_data_from_cloud(self, ems_sn, token):
        """获取未来24小时天气数据"""
        try:
            weather_url = self.base_url + "/ess/web/v1/powerStation/station/solcast/get"
            headers = {"token": token, "Content-Type": "application/json"}
            post_dict = {"registerNo": ems_sn, "day": 1}
            response_data = self.request(
                url=weather_url,
                headers=headers,
                body=post_dict,
            )
            if response_data.get("result") is None:
                app_log.info(f"获取天气数据失败 返回数据 {response_data}")
                return False
            weather_info = {
                "timeList": [],
                "irradiance": {"valueList": []},
                "temperature": {"valueList": []},
            }
            for info in response_data["result"]:
                weather_info["timeList"].append(info["localTime"])
                weather_info["irradiance"]["valueList"].append(float(info["ghi"]))
                weather_info["temperature"]["valueList"].append(float(info["air_temp"]))
            data_length = len(weather_info["irradiance"]["valueList"])
            if weather_info["timeList"] == []:
                weather_info = {}
            else:
                weather_info["irradiance"]["maxValue"] = max(
                    weather_info["irradiance"]["valueList"]
                )
                weather_info["irradiance"]["avgValue"] = round(
                    sum(weather_info["irradiance"]["valueList"]) / data_length, 3
                )
                weather_info["irradiance"]["minValue"] = min(
                    weather_info["irradiance"]["valueList"]
                )

                weather_info["temperature"]["maxValue"] = max(
                    weather_info["temperature"]["valueList"]
                )
                weather_info["temperature"]["avgValue"] = round(
                    sum(weather_info["temperature"]["valueList"]) / data_length, 3
                )
                weather_info["temperature"]["minValue"] = min(
                    weather_info["temperature"]["valueList"]
                )
            app_log.info("获取天气数据成功")
            return weather_info
        except Exception:
            app_log.info(f"获取天气数据失败 异常 {traceback.format_exc()}")
            return False

    def get_electrovalence_data_from_cloud(self, ems_sn, token):
        try:
            price_url = self.base_url + "/powerStation/station/getCurrentElectrovalence"
            response_data = self.request(
                url=price_url,
                body={"registerNo": ems_sn},
                headers={"token": token},
            )
            if response_data.get("result") is None:
                app_log.info(f"获取电价数据失败 返回数据 {response_data}")
                return False
            if ele_price_info := self.parse_electrovalence_data(
                response_data["result"]
            ):
                app_log.info("获取电价数据成功")
                return ele_price_info
            return False
        except Exception:
            app_log.info(f"获取电价数据失败 异常 {traceback.format_exc()}")
            return False

    def parse_electrovalence_data(self, ele_price_data, now_time=None):
        today = datetime.strftime(
            datetime.now() if now_time is None else now_time, "%Y-%m-%d %H:%M:%S"
        )
        ele_price_info = {
            "buy": {},
            "sell": {},
            "date": today,
        }
        currency_code = ele_price_data["currencyCode"]
        transfer_map = {
            "GBP": (100, "p"),
            "EUR": (100, "c"),
            "SEK": (100, "öre"),
        }
        unit = ele_price_data["unit"]
        rate = 1
        if currency_code in transfer_map:
            rate, unit = transfer_map[currency_code]
        ele_price_info["ele_unit"] = f"{unit}/kWh"
        for detail_info in chain(
            ele_price_data["list"],
            ele_price_data.get("tomorrow", []),
        ):
            start_timestamp = detail_info["startTimeStamp"]
            end_timestamp = detail_info["endTimeStamp"]
            if start_timestamp is None or end_timestamp is None:
                continue
            # 处理欧分的情况
            buy_price = (
                round(detail_info["buyPrice"] * rate, 5)
                if detail_info["buyPrice"]
                else None
            )
            sale_price = (
                round(detail_info["salePrice"] * rate, 5)
                if detail_info["salePrice"]
                else None
            )
            while start_timestamp < end_timestamp:  # 每十五分钟存一段
                if buy_price:
                    ele_price_info["buy"][start_timestamp] = buy_price
                if sale_price:
                    ele_price_info["sell"][start_timestamp] = sale_price
                start_timestamp += 15 * 60 * 1000
        return ele_price_info



from typing import Optional, Any

from pydantic import BaseModel

from solax_py_library.smart_scene.types.action import SmartSceneActionExecute


class SmartSceneExecuteResult(BaseModel):
    needLog: bool
    actionType: Optional[SmartSceneActionExecute]
    sceneInstance: Any
    actionIsSuccessful: Optional[bool]



import datetime

from tornado.log import app_log

from solax_py_library.smart_scene.core.condition.base import BaseCondition
from solax_py_library.smart_scene.types.condition import (
    DateConditionItemData,
    DateConditionType,
    ConditionType,
    TimeConditionType,
)
from solax_py_library.utils.time_util import parse_time


class DateCondition(BaseCondition):
    condition_type = ConditionType.date

    def __init__(self, update_value_function, **kwargs):
        super().__init__(update_value_function, **kwargs)

    def meet_func(self, data: DateConditionItemData, ctx):
        if data.childType == DateConditionType.time:
            time_type = data.childData.data[0]
            if time_type == TimeConditionType.time:
                return self.check_time_point(data.childData.data)
            elif time_type == TimeConditionType.timeRange:
                return self.check_start_end_time(data.childData.data)
        return False

    def check_time_point(self, data):
        _, time_point = data
        hour, minute = parse_time(time_point)
        app_log.info(
            f"meet_time: {self.value.get('hour')}, {self.value.get('minute')}, data_value: {time_point}"
        )
        if hour == self.value.get("hour") and minute == self.value.get("minute"):
            return True
        return False

    def check_start_end_time(self, data):
        _, start, end = data

        start_hour, start_minute = parse_time(start)
        end_hour, end_minute = parse_time(end)

        year, month, day, hour, minute = (
            self.value.get(k) for k in ("year", "month", "day", "hour", "minute")
        )

        cur_timestamp = datetime.datetime.strptime(
            f'{year}-{month}-{day} {self.value.get("hour")}:{self.value.get("minute")}',
            "%Y-%m-%d %H:%M",
        )
        start_timestamp = datetime.datetime.strptime(
            f"{year}-{month}-{day} {start_hour}:{start_minute}", "%Y-%m-%d %H:%M"
        )
        end_timestamp = datetime.datetime.strptime(
            f"{year}-{month}-{day} {end_hour}:{end_minute}:59", "%Y-%m-%d %H:%M:%S"
        )

        if end_hour > start_hour and (
            end_hour == start_hour and end_minute > start_minute
        ):
            return True if start_timestamp <= cur_timestamp <= end_timestamp else False
        elif end_hour == start_hour and end_minute == start_minute:
            return True
        else:
            if (
                start_timestamp
                <= cur_timestamp
                <= datetime.datetime.strptime(
                    f"{year}-{month}-{day} 23:59:59", "%Y-%m-%d %H:%M:%S"
                )
            ):
                return True
            if (
                datetime.datetime.strptime(
                    f"{year}-{month}-{day} 00:00", "%Y-%m-%d %H:%M"
                )
                <= cur_timestamp
                <= end_timestamp
            ):
                return True
        return False



from typing import Union, List

from pydantic import BaseModel
from enum import StrEnum


class ModbusPoint(BaseModel):
    starting_address: int
    quantity_of_x: int
    function_code: int


class RegisterDataformat(StrEnum):
    int8 = "int8"
    uint8 = "uint8"
    int16 = "int16"
    uint16 = "uint16"
    int32 = "int32"
    uint32 = "uint32"
    int64 = "int64"
    uint64 = "uint64"
    float = "float"


class ReadModbusPoint(ModbusPoint):
    register_dataformat: RegisterDataformat


class WriteModbusPoint(ModbusPoint):
    output_value: Union[int, List[int], None]



from typing import Dict


INTERNATIONAL_KEY = [
    "self_use",
    "feedin_priority",
    "back_up_mode",
    "manual_mode",
    "forced_charging",
    "forced_discharging",
    "stop_charging_and_discharging",
    "peak_shaving",
    "vpp",
    "power_control_mode",
    "electric_quantity_target_control_mode",
    "soc_target_control_mode",
    "push_power_positive_negative_mode",
    "push_power_zero_mode",
    "self_consume_charge_discharge_mode",
    "self_consume_charge_only_mode",
    "pv_bat_individual_setting_duration_mode",
    "pv_bat_individual_setting_target_soc_mode",
    "rec_name_01",
    "rec_name_02",
    "instruction_01",
    "instruction_02",
    "gt",
    "lt",
    "eq",
    "or",
    "and",
    "once",
    "everyday",
    "weekday",
    "weekend",
    "date",
    "weather",
    "buyingPrice",
    "sellingPrice",
    "systemCondition",
    "cabinet",
    "irradiance",
    "temperature",
    "price",
    "lowerPrice",
    "higherPrice",
    "expensiveHours",
    "cheapestHours",
    "time",
    "duration",
    "systemSoc",
    "systemImportPower",
    "systemExportPower",
    "cabinetSoc",
    "cabinetAlarm",
    "cabinetDo5",
    "if",
    "then",
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
    "tips_alarm",
    "normal_alarm",
    "emergency_alarm",
    "pcs",
    "bms",
    "elm",
    "io",
    "air_conditioner",
    "liquid_cooling_unit",
    "environment_monitoring",
    "ess",
    "DoControl",
    "systemSwitch",
    "exportControl",
    "exportControlOff",
    "importControl",
    "importControlOff",
    "importControl_AELIO",
    "importControl_standby",
    "importControl_discharge",
    "workMode",
    "system",
    "on",
    "off",
    "total",
    "per_phase",
    "electricityPrice",
    "or",
    "and",
    "timeRange",
    "systemLoad",
]


class SmartSceneTranslator:
    _international_map: Dict[str, str] = {key: {} for key in INTERNATIONAL_KEY}

    @classmethod
    def set_internation_map(cls, map_data: Dict[str, str]):
        """Set the internationalization map for the model."""
        cls._international_map = map_data

    @classmethod
    def translate(cls, key: str, lang: str) -> str:
        """Translate a key to the specified language."""
        return cls._international_map.get(key, {}).get(lang, key)


translator = SmartSceneTranslator()



import copy
import struct
from typing import List

format_map = {
    "int8": {"format_str": "bb", "length": 1},
    "uint8": {"format_str": "BB", "length": 1},
    "int16": {"format_str": "h", "length": 1},
    "uint16": {"format_str": "H", "length": 1},
    "int32": {"format_str": "i", "length": 2},
    "uint32": {"format_str": "I", "length": 2},
    "int64": {"format_str": "q", "length": 4},
    "uint64": {"format_str": "Q", "length": 4},
    "float": {"format_str": "f", "length": 1},
}


def unpack(data: List, data_format, reversed=False):
    """
    :param data: 数据字节, 入参均是由modbus读取到的list[uint16]进行转换
    :param data_format:  数据格式
    :param reversed:  是否翻转大小端
    """
    cur_data = copy.deepcopy(data)
    data_format = data_format.lower()
    if data_format not in format_map:
        raise Exception("暂不支持")
    pack_str = ("<" if reversed else ">") + "H" * len(cur_data)
    to_pack_data = struct.pack(pack_str, *cur_data)
    struct_format = ("<" if reversed else ">") + format_map[data_format]["format_str"]
    return struct.unpack(struct_format, to_pack_data)


def pack(value, fmt, order="big"):
    """将10进制的原始值转换为modbus协议需要的精度与类型的值"""
    opt = "<" if order == "little" else ">"
    if fmt not in format_map:
        raise Exception("暂不支持")
    value = int(value)
    ret = struct.pack(f'{opt}{format_map[fmt]["format_str"]}', value)
    ret_list = struct.unpack(f'{opt}{"H" * format_map[fmt]["length"]}', ret)
    return list(ret_list)



import operator
from enum import IntEnum, Enum
from typing import Optional, List, Union, Any

from pydantic import validator, BaseModel, Field, root_validator

from solax_py_library.device.types.alarm import AlarmLevel
from solax_py_library.device.types.device import DeviceType
from solax_py_library.smart_scene.exceptions.smart_scene import SocLimit
from solax_py_library.smart_scene.types.base import translator


class LogicFunc(IntEnum):
    OR = 0
    AND = 1

    def __str__(self):
        return {
            self.OR: "or",
            self.AND: "and",
        }.get(self)


class ConditionFunc(IntEnum):
    GT = 100
    LT = 101
    EQ = 102

    def function(self):
        return {
            ConditionFunc.GT: operator.gt,
            ConditionFunc.LT: operator.lt,
            ConditionFunc.EQ: operator.eq,
        }.get(self)

    def __str__(self):
        return {
            self.GT: "gt",
            self.LT: "lt",
            self.EQ: "eq",
        }.get(self)


class RepeatFunc(IntEnum):
    ONCE = 103
    EVERYDAY = 104
    WEEKDAY = 105
    WEEKEND = 106
    CUSTOM = 107

    def __str__(self):
        return {
            self.ONCE: "once",
            self.EVERYDAY: "everyday",
            self.WEEKDAY: "weekday",
            self.WEEKEND: "weekend",
            self.CUSTOM: "custom",
        }.get(self)


class WeekDay(IntEnum):
    MONDAY = 1
    TUESDAY = 2
    WEDNESDAY = 3
    THURSDAY = 4
    FRIDAY = 5
    SATURDAY = 6
    SUNDAY = 7

    def __str__(self):
        return {
            self.MONDAY: "monday",
            self.TUESDAY: "tuesday",
            self.WEDNESDAY: "wednesday",
            self.THURSDAY: "thursday",
            self.FRIDAY: "friday",
            self.SATURDAY: "saturday",
            self.SUNDAY: "sunday",
        }.get(self)


class ConditionType(str, Enum):
    date = "date"
    weather = "weather"
    buyingPrice = "buyingPrice"
    sellingPrice = "sellingPrice"
    systemCondition = "systemCondition"
    cabinet = "cabinet"


class WeatherConditionType(str, Enum):
    irradiance = "irradiance"
    temperature = "temperature"


class PriceConditionType(str, Enum):
    price = "price"
    lowerPrice = "lowerPrice"
    higherPrice = "higherPrice"
    expensiveHours = "expensiveHours"
    cheapestHours = "cheapestHours"


class DateConditionType(str, Enum):
    time = "time"
    duration = "duration"  # deprecated


class TimeConditionType(IntEnum):
    time = 1
    timeRange = 2

    def __str__(self):
        return {
            self.time: "time",
            self.timeRange: "timeRange",
        }.get(self)


class SystemConditionType(str, Enum):
    systemSoc = "systemSoc"
    systemImportPower = "systemImportPower"  # 买电功率
    systemExportPower = "systemExportPower"  # 馈电功率
    systemLoad = "systemLoad"


class CabinetConditionType(str, Enum):
    cabinetAlarm = "cabinetAlarm"
    cabinetSoc = "cabinetSoc"
    cabinetDo5 = "cabinetDo5"


class SmartSceneUnit(IntEnum):
    PERCENT = 1
    NUM = 2


class ConditionItemChildData(BaseModel):
    function: Optional[ConditionFunc]
    data: List[Any]


class ConditionItemData(BaseModel):
    def check_param(self, ctx):
        ...


class PriceConditionItemData(ConditionItemData):
    childType: PriceConditionType
    childData: ConditionItemChildData

    @validator("childData", always=True)
    def _check_child_data(cls, value, values):
        child_type = values.get("childType")
        if child_type in {
            PriceConditionType.lowerPrice,
            PriceConditionType.higherPrice,
        }:
            assert value.data[0] > 0, ValueError
        elif child_type in {
            PriceConditionType.expensiveHours,
            PriceConditionType.cheapestHours,
        }:
            assert 1 <= value.data[2] <= 24, ValueError
        return value

    def to_text(self, lang, unit):
        data = self.childData.data
        func = self.childData.function
        if self.childType == PriceConditionType.price:
            return translator.translate(self.childType, lang).format(
                translator.translate(str(func), lang),
                f"{data[0]:.5f}".rstrip("0").rstrip("."),
                unit,
            )
        elif self.childType in {
            PriceConditionType.lowerPrice,
            PriceConditionType.higherPrice,
        }:
            return translator.translate(self.childType, lang).format(
                f"{data[0]:.5f}".rstrip("0").rstrip("."), "%" if data[1] == 1 else unit
            )
        elif self.childType in {
            PriceConditionType.expensiveHours,
            PriceConditionType.cheapestHours,
        }:
            return translator.translate(self.childType, lang).format(
                data[0], data[1], data[2]
            )


class SystemConditionItemData(ConditionItemData):
    childType: SystemConditionType
    childData: ConditionItemChildData

    @validator("childData", always=True)
    def _check_child_data(cls, value, values):
        child_type = values.get("childType")
        if child_type in {
            SystemConditionType.systemExportPower,
            SystemConditionType.systemImportPower,
        }:
            assert 0 <= value.data[0] <= 100000, ValueError
            value.data[0] = round(value.data[0], 2)  # 功率保留两位小数
        elif child_type == SystemConditionType.systemLoad:
            assert 0 <= value.data[0] <= 1000, ValueError
            value.data[0] = round(value.data[0], 2)  # 功率保留两位小数
        return value

    def to_text(self, lang, unit):
        data = self.childData.data
        func = self.childData.function
        if self.childType == SystemConditionType.systemSoc:
            return translator.translate(self.childType, lang).format(
                translator.translate(str(func), lang), data[0]
            )
        elif self.childType in {
            SystemConditionType.systemImportPower,
            SystemConditionType.systemExportPower,
        }:
            return translator.translate(self.childType, lang).format(
                translator.translate(str(func), lang), data[0]
            )
        elif self.childType == SystemConditionType.systemLoad:
            return translator.translate(self.childType, lang).format(
                translator.translate(str(func), lang), data[0]
            )

    def check_param(self, ctx):
        soc_low_limit = ctx.get("soc_low_limit", 5)
        if self.childType == SystemConditionType.systemSoc:
            soc = self.childData.data[0]
            if soc < soc_low_limit or soc > 100:
                return SocLimit, {"low_limit": soc_low_limit}


class CabinetConditionItemData(ConditionItemData):
    childType: CabinetConditionType
    childData: ConditionItemChildData

    @validator("childData", always=True)
    def _check_child_data(cls, value, values):
        child_type = values.get("childType")
        if child_type == CabinetConditionType.cabinetAlarm:
            assert value.data[-1] in {
                AlarmLevel.TIPS,
                AlarmLevel.NORMAL,
                AlarmLevel.EMERGENCY,
            }, ValueError
        return value

    def to_text(self, lang, unit):
        data = self.childData.data
        func = self.childData.function
        if self.childType == CabinetConditionType.cabinetSoc:
            return translator.translate(self.childType, lang).format(
                translator.translate(str(func), lang), data[0]
            )
        elif self.childType == CabinetConditionType.cabinetAlarm:
            return translator.translate(self.childType, lang).format(
                ",".join(
                    [
                        translator.translate(str(DeviceType(device_type)), lang)
                        for device_type in data[:-1]
                    ]
                ),
                translator.translate(str(AlarmLevel(data[-1])), lang),
            )
        elif self.childType == CabinetConditionType.cabinetDo5:
            return translator.translate(self.childType, lang)

    def check_param(self, ctx):
        soc_low_limit = ctx.get("soc_low_limit", 5)
        if self.childType == CabinetConditionType.cabinetSoc:
            soc = self.childData.data[0]
            if soc < soc_low_limit or soc > 100:
                return SocLimit, {"low_limit": soc_low_limit}


class DateConditionItemData(ConditionItemData):
    childType: DateConditionType
    childData: ConditionItemChildData

    @validator("childData", always=True)
    def _check_child_data(cls, value, values):
        child_type = values.get("childType")
        data = value.data
        if child_type == DateConditionType.time:
            time_type = data[0]
            if time_type == TimeConditionType.time:
                assert isinstance(data[1], str), ValueError
            elif time_type == TimeConditionType.timeRange:
                assert isinstance(data[1], str), ValueError
                assert isinstance(data[2], str), ValueError
        # elif child_type == DateConditionType.duration:
        #     assert isinstance(data[0], int), ValueError
        return value

    def to_text(self, lang, unit):
        if self.childType == DateConditionType.duration:
            return translator.translate(self.childType, lang).format(
                self.childData.data[0]
            )
        elif self.childType == DateConditionType.time:
            if self.childData.data[0] == TimeConditionType.time:
                return (
                    translator.translate(
                        str(TimeConditionType(self.childData.data[0])), lang
                    )
                    + self.childData.data[1]
                )
            elif self.childData.data[0] == TimeConditionType.timeRange:
                return translator.translate(
                    str(TimeConditionType(self.childData.data[0])), lang
                ).format(self.childData.data[1], self.childData.data[2])


class WeatherConditionItemData(ConditionItemData):
    childType: WeatherConditionType
    childData: ConditionItemChildData

    @validator("childData", always=True)
    def _check_child_data(cls, value, values):
        child_type = values.get("childType")
        if child_type == WeatherConditionType.irradiance:
            assert value.data[0] > 0, ValueError
            assert 0 <= value.data[1] <= 24, ValueError
        return value

    def to_text(self, lang, unit):
        func = self.childData.function
        data = self.childData.data
        if self.childType == WeatherConditionType.irradiance:
            return translator.translate(self.childType, lang).format(
                translator.translate(str(func), lang), data[0], data[1]
            )
        else:
            return translator.translate(self.childType, lang).format(
                translator.translate(str(func), lang), data[0]
            )


class ConditionItem(BaseModel):
    type: ConditionType
    cabinet: Optional[List[str]]
    data: List[
        Union[
            DateConditionItemData,
            WeatherConditionItemData,
            PriceConditionItemData,
            SystemConditionItemData,
            CabinetConditionItemData,
        ]
    ]

    @validator("cabinet")
    def _check_cabinet(cls, value, values):
        condition_type = values.get("type")
        if condition_type == ConditionType.cabinet:
            assert value, "cabinet is None"
        return value

    def to_text(self, lang, unit):
        if self.type != ConditionType.cabinet:
            return {self.type: [d.to_text(lang, unit) for d in self.data]}
        elif self.type == ConditionType.cabinet:
            cabinet_sns = ",".join(self.cabinet)
            return {
                self.type: [d.to_text(lang, unit) for d in self.data] + [cabinet_sns]
            }


class SmartSceneCondition(BaseModel):
    operation: LogicFunc
    durationTime: Optional[int] = Field(description="时序时间(s)")
    value: List[ConditionItem]

    @root_validator(pre=True)
    def _check_root(cls, values):
        operation = values.get("operation")
        if operation == LogicFunc.OR:
            values["durationTime"] = None
        return values

    def to_text(self, lang, unit):
        ret = {
            "operation": [translator.translate(str(self.operation), lang)],
        }
        if self.durationTime:
            ret["durationTime"] = [
                translator.translate("duration", lang).format(self.durationTime)
            ]
        for v in self.value:
            ret.update(v.to_text(lang, unit))
        return ret

    # @deprecated
    def get_duration_info(self):
        for item in self.value:
            if item.type != ConditionType.date:
                continue
            for date_item in item.data:
                if date_item.childType == DateConditionType.duration:
                    return (date_item.childData.data[0] // 10) + 1
        return 1



import asyncio
from decimal import Decimal, ROUND_HALF_UP


def retry(max_attempts=3, delay=0.5, assign_exception=Exception):
    def decorator(func):
        async def wrapper(*args, **kwargs):
            attempts = 0
            while attempts < max_attempts:
                try:
                    return await func(*args, **kwargs)
                except Exception as e:
                    print(f"尝试 {attempts+1} 失败: {e}")
                    await asyncio.sleep(delay)
                    attempts += 1
            raise assign_exception(f"操作失败 {max_attempts} 次")

        return wrapper

    return decorator


def round_value(value, decimal_places=0):
    """
    实现四舍五入 代替round函数 遇到5向上取整
    @param value: 传入数值
    @param decimal_places: 保留几位小数
    @return:
    """
    if isinstance(value, int):
        return value
    else:
        # 优化小数位 Decimal函数太耗时 曲线数据直接使用round函数
        # rounded_value = ("{:.%df}" % decimal_places).format(value)
        rounded_value = Decimal(str(value)).quantize(
            Decimal("1e-" + str(decimal_places)), rounding=ROUND_HALF_UP
        )
        return float(rounded_value)



import json
import traceback
from datetime import datetime
from typing import List, Set

from solax_py_library.smart_scene.core.condition import BaseCondition
from solax_py_library.smart_scene.types.action import (
    SmartSceneAction,
    SmartSceneActionExecute,
)
from solax_py_library.smart_scene.types.condition import (
    ConditionType,
    RepeatFunc,
    LogicFunc,
    SmartSceneCondition,
    DateConditionType,
    ConditionItem,
)
from solax_py_library.smart_scene.types.smart_scene_content import (
    SmartSceneContent,
    SmartSceneOtherInfo,
)
from tornado.log import app_log

from solax_py_library.smart_scene.types.smart_scene_result import (
    SmartSceneExecuteResult,
)


class SmartSceneRunner:
    def __init__(self, condition_classes, action_classes):
        self.condition_class_map = {}
        self.action_class_map = {}

        for condition_class in condition_classes:
            if not isinstance(condition_class, BaseCondition):
                raise ValueError(
                    f"condition({condition_class}) is not derived from BaseCondition"
                )
            self.condition_class_map[condition_class.condition_type] = condition_class

        for action_class in action_classes:
            self.action_class_map[action_class.action_type] = action_class

    @staticmethod
    def _handle_repeat_info(scene_content: SmartSceneContent, once_flag):
        # once_flag: 是否执行过
        if scene_content.repeatFunction == RepeatFunc.ONCE:
            return False if once_flag else True
        week_list = scene_content.get_weekday_index()
        if datetime.now().weekday() + 1 in week_list:
            return True
        return False

    def smart_scene_handle(self, scene_instance):
        other_info = SmartSceneOtherInfo.parse_raw(scene_instance.other)
        try:
            scene_data = SmartSceneContent.parse_raw(scene_instance.content)

            # repeat，是否执行
            if not self._handle_repeat_info(scene_data, other_info.once_flag):
                app_log.info(
                    f"场景名称: {scene_data.name}, 场景ID: {scene_instance.scene_id} 不符合重复条件，不执行"
                )
                return
            scene_id = scene_instance.scene_id
            app_log.info(
                f"场景名称: {scene_data.name}, 场景ID: {scene_id}, 判断类型: {scene_data.conditions.operation}:"
            )

            # 条件判断
            result = self._handle_conditions(
                scene_data.get_duration_times(),
                scene_data.conditions.operation,
                scene_data.conditions,
                other_info,
            )

            # 动作执行
            need_log = False
            new_exec_number = scene_instance.exec_number
            if result:
                action_is_successful = self._handle_action(
                    scene_id, scene_data.thenActions, log_prefix="THEN"
                )
                new_exec_number = 1
                # 何种情况下，需要记录log
                if other_info.is_log != SmartSceneActionExecute.THEN or (
                    other_info.is_log == SmartSceneActionExecute.THEN
                    and action_is_successful != other_info.action_is_successful
                ):
                    need_log = True
                if scene_data.repeatFunction == RepeatFunc.ONCE:
                    scene_instance.switch = False
                other_info.set_then_result(action_is_successful)
            else:
                if scene_data.elseThenActions:
                    if scene_instance.exec_number == 1:
                        action_is_successful = self._handle_action(
                            scene_id, scene_data.elseThenActions, log_prefix="ELSE THEN"
                        )
                        new_exec_number = 0
                        if other_info.is_log != SmartSceneActionExecute.ELSE_THEN:
                            need_log = True
                        other_info.set_else_then_result(action_is_successful)
                else:
                    new_exec_number = 0
                    other_info.set_none_action()

            app_log.info(f"{scene_id} 执行完毕\n")
            scene_instance.exec_number = new_exec_number
            scene_instance.other = json.dumps(other_info.dict())
            return SmartSceneExecuteResult(
                actionType=other_info.is_log,
                sceneInstance=scene_instance,
                actionIsSuccessful=other_info.action_is_successful,
                needLog=need_log,
            )
        except Exception:
            app_log.error(
                f"{scene_instance.scene_id} 执行智能场景异常 {traceback.format_exc()}"
            )

    def _handle_action(
        self, scene_id, actions: List[SmartSceneAction], log_prefix="THEN"
    ):
        if not actions:
            return
        action_is_successful = True
        for action_info in actions:
            action_class = self.action_class_map[action_info.type]
            for child_info in action_info.data:
                ret = action_class.do_func(scene_id=scene_id, data=child_info)
                app_log.info(f"{log_prefix}条件 {child_info} 执行结果: {ret}")
                action_is_successful &= ret
        return action_is_successful

    def _handle_conditions(
        self,
        need_duration_times,
        operation: LogicFunc,
        conditions: SmartSceneCondition,
        other_info: SmartSceneOtherInfo,
    ):
        if not self._check_conditions(operation, conditions):
            other_info.duration_times = 0
            return False
        if other_info.duration_times < need_duration_times:
            other_info.duration_times += 1
        app_log.info(
            f"need times: {need_duration_times}, current times: {other_info.duration_times}"
        )
        return other_info.duration_times >= need_duration_times

    def _check_conditions(self, operation, conditions):
        ret_list = []
        for cond in conditions.value:
            parent_type = cond.type  # 父条件类型 (日期、天气、电价)
            condition_class = self.condition_class_map[parent_type]
            ctx = self._build_condition_ctx(cond)
            for child_info in cond.data:
                if child_info.childType == DateConditionType.duration:
                    continue
                ret = condition_class.meet_func(data=child_info, ctx=ctx)
                app_log.info(f"IF条件 {child_info} 判断结果: {ret}")
                ret_list.append(ret)
                # 如果该条件判断不满足，并且客户设置为and,则无需继续判断，返回失败
                if not ret and operation == LogicFunc.AND:
                    return False
                # 如果条件判断满足，并且客户设置为or,则无需继续判断，返回成功
                if ret and operation == LogicFunc.OR:
                    return True
        if (operation == LogicFunc.AND and all(ret_list)) or (
            operation == LogicFunc.OR and any(ret_list)
        ):
            return True
        return False

    @staticmethod
    def _build_condition_ctx(condition_info: ConditionItem):
        ctx = {}
        if condition_info.type == ConditionType.cabinet:
            ctx["cabinet"] = condition_info.cabinet
        return ctx

    def update_all_condition_data(self, condition_types: Set[ConditionType]):
        for cond_type in condition_types:
            self.condition_class_map[cond_type].update_value()



from tornado.log import app_log

from solax_py_library.smart_scene.core.condition.base import BaseCondition
from solax_py_library.smart_scene.types.condition import (
    SystemConditionItemData,
    SystemConditionType,
    ConditionType,
)


class SystemCondition(BaseCondition):
    condition_type = ConditionType.systemCondition

    def __init__(self, update_value_function, **kwargs):
        super().__init__(update_value_function, **kwargs)

    def meet_func(self, data: SystemConditionItemData, ctx):
        if not self.value:
            return False
        child_data = data.childData
        function_value = child_data.function
        compare_value = None
        if data.childType == SystemConditionType.systemExportPower:
            compare_value = self.value.get("grid_active_power")
            if compare_value < 0:
                return False
            app_log.info(
                f"meet_system_export_power: {compare_value}, data_value: {child_data.data[0]}"
            )
        elif data.childType == SystemConditionType.systemImportPower:
            compare_value = self.value.get("grid_active_power")
            if compare_value > 0:
                return False
            app_log.info(
                f"meet_system_import_power: {compare_value}, data_value: {child_data.data[0]}"
            )
        elif data.childType == SystemConditionType.systemSoc:
            compare_value = self.value.get("system_soc")
            app_log.info(
                f"meet_system_soc: {compare_value}, data_value: {child_data.data[0]}"
            )
        elif data.childType == SystemConditionType.systemLoad:
            compare_value = self.value.get("system_load")
            app_log.info(
                f"meet_system_load: {compare_value}, data_value: {child_data.data[0]}"
            )
        if compare_value is None:
            return False
        return function_value.function()(abs(compare_value), child_data.data[0])



from typing import List, Optional, Dict
from modbus_tk import defines

from pydantic import BaseModel
from enum import IntEnum

from solax_py_library.device.types.modbus_point import ReadModbusPoint, WriteModbusPoint


class InverterModel(IntEnum):
    ...

    def __str__(self):
        ...


class InverterPoint(ReadModbusPoint):
    description: Optional[str]
    point_key: str
    point_id: int
    accuracy: Optional[float]
    function_code: int = defines.READ_INPUT_REGISTERS


class InverterVersionConfig(BaseModel):
    software: ReadModbusPoint
    hardware: ReadModbusPoint


class InverterSystemConfig(BaseModel):
    sn: ReadModbusPoint
    shutdown: WriteModbusPoint
    boot_up: WriteModbusPoint
    time: WriteModbusPoint
    version: InverterVersionConfig


class InverterConfig(BaseModel):
    real_data: List[InverterPoint]
    other_data: InverterSystemConfig
    error_map: Dict[int, int]
